@@ -69,6 +69,7 @@ lg.add('UNTERMINATED_STRING', r"[\"\'].*")
 lg.add('NUMBER', r'-?[0-9]+(?:\.[0-9]+)?')
 lg.add('NAME', r'\&?[_a-zA-Z$][-_a-zA-Z0-9]*')
 lg.add('OPPAREN', r'" + operator_regex_str + "(?=\()')
+lg.add('PIPELINE_BIND', r'\|>\?')
 lg.add('PIPELINE', r'\|>')
 lg.add('BAR', r'\|')
 lg.add('LBRACK', r'\[')
@@ -78,6 +79,7 @@ lg.add('RBRACE', r'\}')
 lg.add('LPAREN', r'\(')
 lg.add('RPAREN', r'\)')
 lg.add('DOT', r'\.')
+lg.add('BANG', r'!')
 lg.add('PERCENT', r'%')
 lg.add('COMMA', r',')
 lg.add('THINARROW', r'->')
@@ -95,8 +97,6 @@ lg.add('OPNEQ', r'!=')
 lg.add('OPLT', r'<')
 lg.add('OPGT', r'>')
 
-lg.add('BANG', r'!')
-
 lg.add('EQUALS', r'=')
 lg.add('SEMI', r';')
 lg.add('AT', r'@')
@@ -130,6 +130,7 @@ klg.add('THENCOLON', r'^then:$')
 klg.add('ELSE', r'^else$')
 klg.add('ELSEIF', r'^elif$')
 klg.add('MATCH', r'^match$')
+klg.add('RECEIVE', r'^receive$')
 klg.add('OF', r'^of$')
 klg.add('RECORD', r'^record$')
 klg.add('DATA', r'^data$')
@@ -150,10 +151,10 @@ klg.add('NOT', r'^not$')
 
 pg = ParserGenerator(['NUMBER', 'OPPLUS', 'OPMINUS', 'OPTIMES', 'OPDIV', 'OPLEQ', 'OPGEQ', 'OPEQ', 'OPNEQ',
                       'OPLT', 'OPGT', 'OPAND', 'OPOR', 'OPIS', 'NOT', 'NEWLINE', 'PERCENT', 'EXPORT',
-                      'LPAREN', 'RPAREN', 'TRUE', 'FALSE', 'DQUOTE_STR', 'SQUOTE_STR', 'AT',
+                      'LPAREN', 'RPAREN', 'TRUE', 'FALSE', 'DQUOTE_STR', 'SQUOTE_STR', 'AT', 'BANG',
                       'NAME', 'EQUALS', 'IF', 'ELSEIF', 'ELSE', 'COLON', 'SEMI', 'DATA', 'IMPORT', 'REQUIRE',
-                      'LBRACK', 'RBRACK', 'COMMA', 'DEF', 'DOC', 'CALET', 'PIPELINE', 'RETURN',
-                      'LBRACE', 'RBRACE', 'MATCH', 'DEFM', 'RECORD', 'AMP', 'FN', 'THINARROW',
+                      'LBRACK', 'RBRACK', 'COMMA', 'DEF', 'DOC', 'CALET', 'PIPELINE', 'PIPELINE_BIND', 'RETURN',
+                      'LBRACE', 'RBRACE', 'MATCH', 'DEFM', 'RECORD', 'AMP', 'FN', 'THINARROW', 'RECEIVE',
                       'YIELD', 'FROM', 'FOR', 'IN', 'DOT', 'INDENT', 'DEDENT', 'TRY', 'FINALLY', 'EXCEPT',
                       'MODULE', 'AS', 'RAISE'],
                      precedence=[('left', ['EQUALS']),
@@ -178,7 +179,6 @@ def block(p):
 
 @pg.production('stmts : stmts stmt')
 def stmts_b(p):
-    # if issequence_except_str(p[1]):
     if p[1] is None:
         return p[0]
     else:
@@ -376,12 +376,14 @@ def binding(p):
 @pg.production('expr : dict_expr')
 @pg.production('expr : tuple_expr')
 @pg.production('expr : match_expr')
+@pg.production('expr : receive_expr')
 @pg.production('expr : yield_expr')
 @pg.production('expr : yield_from_expr')
 @pg.production('expr : for_expr')
 @pg.production('expr : block_expr')
 @pg.production('expr : defm_expr')
 @pg.production('expr : dot_expr')
+@pg.production('expr : send_msg_expr')
 @pg.production('expr : id_expr')
 # @pg.production('expr : app_nc_expr')
 def expr(p):
@@ -525,6 +527,11 @@ def dot_expr(p):
     return [Symbol('getattr'), p[0], p[2].getstr()]
 
 
+@pg.production('send_msg_expr : expr BANG expr')
+def dot_expr(p):
+    return [Symbol('send'), p[2], p[0]]
+
+
 @pg.production('for_expr : LBRACK binop_expr FOR pattern IN binop_expr RBRACK')  # TODO
 def for_expr(p):
     pattern = p[3]
@@ -752,6 +759,14 @@ def right_app_expr(p):
     return [right_app_fun_expr] + app_args + [expr]
 
 
+@pg.production('right_app_expr : expr PIPELINE_BIND right_app_fun_expr app_args')
+def right_app_bind_expr(p):
+    expr, _, right_app_fun_expr, app_args = p
+    return [Symbol('if'), [Symbol('is'), expr, Symbol('None')],
+            Symbol('None'),
+            [right_app_fun_expr] + app_args + [expr]]
+
+
 @pg.production('right_app_fun_expr : id_expr')
 def right_app_fun_expr(p):
     return p[0]
@@ -841,6 +856,11 @@ def pattern(p):
     return p[0]
 
 
+@pg.production('receive_expr : RECEIVE COLON NEWLINE INDENT case_branches DEDENT')
+def case(p):
+    return [Symbol('match'), [Symbol('recv'), [Symbol('self')]]] + p[4]
+
+
 @pg.production('record_expr : RECORD id_expr')
 def record_expr(p):
     return [Symbol('record'), p[1], []]
